from __future__ import division
from __future__ import print_function

import argparse
from copy import deepcopy
import json
import numpy as np
from numpy.polynomial import Polynomial
import pkgutil
import sys
from uuid import uuid4

from numina.array.display.polfit_residuals import polfit_residuals
from numina.array.display.ximshow import ximshow_file
from numina.array.display.pause_debugplot import pause_debugplot


def assign_boxes_to_fibers(insmode):
    """Read boxes in configuration file and assign values to fibid

    Parameters
    ----------
    insmode : string
        Value of the INSMODE keyword: 'LCB' or 'MOS'.

    Returns
    -------
    fibid_with_box : list of strings
        List with string label that contains both the fibid and the
        box name.

    """
    sdum = pkgutil.get_data(
        'megaradrp.instrument.configs',
        'component-86a6e968-8d3d-456f-89f8-09ff0c7f7c57.json'
    )
    dictdum = json.loads(sdum)
    pseudo_slit_config = \
        dictdum['configurations']['boxes']['values'][insmode]

    fibid_with_box = []
    n1 = 1
    list_to_print = []
    for dumbox in pseudo_slit_config:
        nfibers = dumbox['nfibers']
        name = dumbox['name']
        n2 = n1 + nfibers
        fibid_with_box += \
            ["{}  [{}]".format(val1, val2)
             for val1, val2 in zip(range(n1, n2), [name] * nfibers)]
        dumstr ='Box {:>2},  fibers {:3d} - {:3d}'.format(name, n1, n2 - 1)
        list_to_print.append(dumstr)
        n1 = n2
    print('\n* Fiber description for INSMODE=' + insmode)
    for dumstr in reversed(list_to_print):
        print(dumstr)
    print('---------------------------------')

    return fibid_with_box


def plot_trace(ax, coeff, xmin, xmax, ix_offset,
               rawimage, fibids, fiblabel, colour):
    if xmin == xmax == 0:
        num = 4096
        xp = np.linspace(start=1, stop=4096, num=num)
    else:
        num = int(float(xmax - xmin + 1) + 0.5)
        xp = np.linspace(start=xmin, stop=xmax, num=num)
    ypol = Polynomial(coeff)
    yp = ypol(xp)
    if rawimage:
        lcut = (yp > 2056.5)
        yp[lcut] += 100
    ax.plot(xp + ix_offset, yp + 1, color=colour, linestyle='dotted')
    if fibids:
        if xmin == xmax == 0:
            xmidpoint = 2048
        else:
            xmidpoint = (xmin+xmax)/2
        ax.text(xmidpoint, yp[int(num / 2)], fiblabel, fontsize=6,
                bbox=dict(boxstyle="round,pad=0.1", fc="white", ec="grey", ),
                color=colour, fontweight='bold', backgroundcolor='white',
                ha='center')


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: overplot traces"
    )
    # positional parameters
    parser.add_argument("fits_file",
                        help="FITS image containing the spectra",
                        type=argparse.FileType('r'))
    parser.add_argument("traces_file",
                        help="JSON file with fiber traces",
                        type=argparse.FileType('r'))
    # optional parameters
    parser.add_argument("--rawimage",
                        help="FITS file is a RAW image (otherwise trimmed "
                             "image is assumed)",
                        action="store_true")
    parser.add_argument("--global_offset",
                        help="Global offset polynomial coefficients "
                             "(+upwards, -downwards)")
    parser.add_argument("--fibids",
                        help="Display fiber identification number",
                        action="store_true")
    parser.add_argument("--verbose",
                        help="Enhance verbosity",
                        action="store_true")
    parser.add_argument("--healing",
                        help="JSON healing file to improve traces",
                        type=argparse.FileType('r'))
    parser.add_argument("--updated_traces",
                        help="JSON file with modified fiber traces",
                        type=argparse.FileType('w'))
    parser.add_argument("--z1z2",
                        help="tuple z1,z2, minmax or None (use zscale)")
    parser.add_argument("--bbox",
                        help="bounding box tuple: nc1,nc2,ns1,ns2")
    parser.add_argument("--keystitle",
                        help="tuple of FITS keywords.format: " +
                             "key1,key2,...keyn.'format'")
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy",
                        default="0,0,640,480")
    parser.add_argument("--pdffile",
                        help="ouput PDF file name",
                        type=argparse.FileType('w'))
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # global_offset in command line
    if args.global_offset is None:
        args_global_offset = [0.0]
    else:
        args_global_offset = [float(dum) for dum in
                              str(args.global_offset).split(",")]

    # read pdffile
    if args.pdffile is not None:
        from matplotlib.backends.backend_pdf import PdfPages
        pdf = PdfPages(args.pdffile.name)
    else:
        pdf = None

    ax = ximshow_file(args.fits_file.name,
                      args_cbar_orientation='vertical',
                      args_z1z2=args.z1z2,
                      args_bbox=args.bbox,
                      args_keystitle=args.keystitle,
                      args_geometry=args.geometry,
                      pdf=pdf,
                      show=False)

    # trace offsets for RAW images
    if args.rawimage:
        ix_offset = 51
    else:
        ix_offset = 1

    # read and display traces from JSON file
    bigdict = json.loads(open(args.traces_file.name).read())
    insmode = bigdict['tags']['insmode']
    fibid_with_box = assign_boxes_to_fibers(insmode)
    total_fibers = bigdict['total_fibers']
    if total_fibers != len(fibid_with_box):
        raise ValueError('Mismatch between number of fibers and '
                         'expected number from account from boxes')
    if 'global_offset' in bigdict.keys():
        global_offset = bigdict['global_offset']
        if args_global_offset != [0.0] and global_offset != [0.0]:
            raise ValueError('global_offset != 0 argument cannot be employed '
                             'when global_offset != 0 in JSON file')
        elif args_global_offset != [0.0]:
            global_offset = args_global_offset
    else:
        global_offset = args_global_offset
    print('>>> Using global_offset:', global_offset)
    pol_global_offset = np.polynomial.Polynomial(global_offset)
    if 'ref_column' in bigdict.keys():
        ref_column = bigdict['ref_column']
    else:
        ref_column = 2000
    for fiberdict in bigdict['contents']:
        fibid = fiberdict['fibid']
        fiblabel = fibid_with_box[fibid - 1]
        start = fiberdict['start']
        stop = fiberdict['stop']
        coeff = np.array(fiberdict['fitparms'])
        # skip fibers without trace
        if len(coeff) > 0:
            pol_trace = np.polynomial.Polynomial(coeff)
            y_at_ref_column = pol_trace(ref_column)
            correction = pol_global_offset(y_at_ref_column)
            coeff[0] += correction
            # update values in bigdict (JSON structure)
            bigdict['contents'][fibid-1]['fitparms'] = coeff.tolist()
            plot_trace(ax, coeff, start, stop, ix_offset, args.rawimage,
                       args.fibids, fiblabel, colour='blue')
        else:
            print('Warning ---> Missing fiber:', fibid_with_box[fibid - 1])

    # if present, read healing JSON file
    if args.healing is not None:
        healdict = json.loads(open(args.healing.name).read())
        list_operations = healdict['operations']
        for operation in list_operations:

            if operation['description'] == 'vertical_shift_in_pixels':
                if 'fibid_list' in operation.keys():
                    fibid_list = operation['fibid_list']
                else:
                    fibid_ini = operation['fibid_ini']
                    fibid_end = operation['fibid_end']
                    fibid_list = range(fibid_ini, fibid_end + 1)
                for fibid in fibid_list:
                    if fibid < 1 or fibid > total_fibers:
                        raise ValueError('fibid number outside valid range')
                    fiblabel = fibid_with_box[fibid - 1]
                    coeff = np.array(
                        bigdict['contents'][fibid - 1]['fitparms']
                    )
                    if len(coeff) > 0:
                        if args.verbose:
                            print('(vertical_shift_in_pixels) fibid:',
                                  fiblabel)
                        vshift = operation['vshift']
                        coeff[0] += vshift
                        bigdict['contents'][fibid - 1]['fitparms'] = \
                            coeff.tolist()
                        start = bigdict['contents'][fibid - 1]['start']
                        stop = bigdict['contents'][fibid - 1]['stop']
                        plot_trace(ax, coeff, start, stop, ix_offset,
                                   args.rawimage, True, fiblabel,
                                   colour='green')
                    else:
                        print('(vertical_shift_in_pixels SKIPPED) fibid:',
                              fiblabel)

            elif operation['description'] == 'duplicate_trace':
                fibid_original = operation['fibid_original']
                if fibid_original < 1 or fibid_original > total_fibers:
                    raise ValueError(
                        'fibid_original number outside valid range'
                    )
                fibid_duplicated = operation['fibid_duplicated']
                if fibid_duplicated < 1 or fibid_duplicated > total_fibers:
                    raise ValueError(
                        'fibid_duplicated number outside valid range'
                    )
                fiblabel_original = fibid_with_box[fibid_original - 1]
                fiblabel_duplicated = fibid_with_box[fibid_duplicated - 1]
                coeff = np.array(
                    bigdict['contents'][fibid_original - 1]['fitparms']
                )
                if len(coeff) > 0:
                    if args.verbose:
                        print('(duplicated_trace) fibids:',
                              fiblabel_original, '-->', fiblabel_duplicated)
                    vshift = operation['vshift']
                    coeff[0] += vshift
                    bigdict['contents'][fibid_duplicated - 1]['fitparms'] = \
                        coeff.tolist()
                    start = bigdict['contents'][fibid_original - 1]['start']
                    stop = bigdict['contents'][fibid_original - 1]['stop']
                    bigdict['contents'][fibid_duplicated - 1]['start'] = start
                    bigdict['contents'][fibid_duplicated - 1]['stop'] = stop
                    plot_trace(ax, coeff, start, stop, ix_offset,
                               args.rawimage, True, fiblabel_duplicated,
                               colour='green')
                else:
                    print('(duplicated_trace SKIPPED) fibids:',
                          fiblabel_original, '-->', fiblabel_duplicated)

            elif operation['description'] == 'extrapolation':
                if 'fibid_list' in operation.keys():
                    fibid_list = operation['fibid_list']
                else:
                    fibid_ini = operation['fibid_ini']
                    fibid_end = operation['fibid_end']
                    fibid_list = range(fibid_ini, fibid_end + 1)
                for fibid in fibid_list:
                    if fibid < 1 or fibid > total_fibers:
                        raise ValueError('fibid number outside valid range')
                    fiblabel = fibid_with_box[fibid - 1]
                    coeff = np.array(
                        bigdict['contents'][fibid - 1]['fitparms']
                    )
                    if len(coeff) > 0:
                        if args.verbose:
                            print('(extrapolation) fibid:', fiblabel)
                        # update values in bigdict (JSON structure)
                        start = operation['start']
                        stop = operation['stop']
                        start_orig = bigdict['contents'][fibid - 1]['start']
                        stop_orig = bigdict['contents'][fibid - 1]['stop']
                        bigdict['contents'][fibid - 1]['start'] = start
                        bigdict['contents'][fibid - 1]['stop'] = stop
                        if start < start_orig:
                            plot_trace(ax, coeff, start, start_orig,
                                       ix_offset,
                                       args.rawimage, True, fiblabel,
                                       colour='green')
                        if stop_orig < stop:
                            plot_trace(ax, coeff, stop_orig, stop,
                                       ix_offset,
                                       args.rawimage, True, fiblabel,
                                       colour='green')
                        if start_orig <= start <= stop <= stop_orig:
                            plot_trace(ax, coeff, start, stop,
                                       ix_offset,
                                       args.rawimage, True, fiblabel,
                                       colour='green')
                    else:
                        print('(extrapolation SKIPPED) fibid:', fiblabel)

            elif operation['description'] == 'fit_through_user_points':
                fibid = operation['fibid']
                fiblabel = fibid_with_box[fibid - 1]
                if args.verbose:
                    print('(fit through user points) fibid:', fiblabel)
                poldeg = operation['poldeg']
                start = operation['start']
                stop = operation['stop']
                xfit = []
                yfit = []
                for userpoint in operation['user_points']:
                    # assume x, y coordinates in JSON file are given in
                    # image coordinates, starting at (1,1) in the lower
                    # left corner
                    xdum = userpoint['x'] - 1  # use np.array coordinates
                    ydum = userpoint['y'] - 1  # use np.array coordinates
                    xfit.append(xdum)
                    yfit.append(ydum)
                xfit = np.array(xfit)
                yfit = np.array(yfit)
                if len(xfit) <= poldeg:
                    raise ValueError('Insufficient number of points to fit'
                                     ' polynomial')
                poly, residum = polfit_residuals(xfit, yfit, poldeg)
                coeff = poly.coef
                plot_trace(ax, coeff, start, stop, ix_offset,
                           args.rawimage, args.fibids, fiblabel,
                           colour='green')
                bigdict['contents'][fibid - 1]['start'] = start
                bigdict['contents'][fibid - 1]['stop'] = stop
                bigdict['contents'][fibid - 1]['fitparms'] = coeff.tolist()

            elif operation['description'] == \
                    'extrapolation_through_user_points':
                fibid = operation['fibid']
                fiblabel = fibid_with_box[fibid - 1]
                if args.verbose:
                    print('(extrapolation_through_user_points):', fiblabel)
                start_reuse = operation['start_reuse']
                stop_reuse = operation['stop_reuse']
                resampling = operation['resampling']
                poldeg = operation['poldeg']
                start = operation['start']
                stop = operation['stop']
                coeff = bigdict['contents'][fibid - 1]['fitparms']
                xfit = np.linspace(start_reuse, stop_reuse, num=resampling)
                poly = np.polynomial.Polynomial(coeff)
                yfit = poly(xfit)
                for userpoint in operation['user_points']:
                    # assume x, y coordinates in JSON file are given in
                    # image coordinates, starting at (1,1) in the lower
                    # left corner
                    xdum = userpoint['x'] - 1  # use np.array coordinates
                    ydum = userpoint['y'] - 1  # use np.array coordinates
                    xfit = np.concatenate((xfit, np.array([xdum])))
                    yfit = np.concatenate((yfit, np.array([ydum])))
                poly, residum = polfit_residuals(xfit, yfit, poldeg)
                coeff = poly.coef
                if start < start_reuse:
                    plot_trace(ax, coeff, start, start_reuse, ix_offset,
                               args.rawimage, args.fibids, fiblabel,
                               colour='green')
                if stop_reuse < stop:
                    plot_trace(ax, coeff, stop_reuse, stop, ix_offset,
                               args.rawimage, args.fibids, fiblabel,
                               colour='green')
                bigdict['contents'][fibid - 1]['start'] = start
                bigdict['contents'][fibid - 1]['stop'] = stop
                bigdict['contents'][fibid - 1]['fitparms'] = coeff.tolist()

            elif operation['description'] == 'sandwich':
                fibid = operation['fibid']
                fiblabel = fibid_with_box[fibid - 1]
                if args.verbose:
                    print('(sandwich) fibid:', fiblabel)
                fraction = operation['fraction']
                nf1, nf2 = operation['neighbours']
                start = operation['start']
                stop = operation['stop']
                tmpf1 = bigdict['contents'][nf1 - 1]
                tmpf2 = bigdict['contents'][nf2 - 1]
                if nf1 != tmpf1['fibid'] or nf2 != tmpf2['fibid']:
                    raise ValueError(
                        "Unexpected fiber numbers in neighbours"
                    )
                coefff1 = np.array(tmpf1['fitparms'])
                coefff2 = np.array(tmpf2['fitparms'])
                coeff = coefff1 + fraction * (coefff2 - coefff1)
                plot_trace(ax, coeff, start, stop, ix_offset,
                           args.rawimage, args.fibids,
                           fiblabel, colour='green')
                # update values in bigdict (JSON structure)
                bigdict['contents'][fibid - 1]['start'] = start
                bigdict['contents'][fibid - 1]['stop'] = stop
                bigdict['contents'][fibid - 1][
                    'fitparms'] = coeff.tolist()
                if fibid in bigdict['error_fitting']:
                    bigdict['error_fitting'].remove(fibid)

            elif operation['description'] == 'renumber_fibids_within_box':
                fibid_ini = operation['fibid_ini']
                fibid_end = operation['fibid_end']
                box_ini = fibid_with_box[fibid_ini - 1][4:]
                box_end = fibid_with_box[fibid_end - 1][4:]
                if box_ini != box_end:
                    print('ERROR: box_ini={}, box_end={}'.format(box_ini,
                                                                 box_end))
                    raise ValueError('fibid_ini and fibid_end correspond to '
                                     'different fiber boxes')
                fibid_shift = operation['fibid_shift']
                if fibid_shift in [-1, 1]:
                    if fibid_shift == -1:
                        i_start = fibid_ini
                        i_stop = fibid_end + 1
                        i_step = 1
                    else:
                        i_start = fibid_end
                        i_stop = fibid_ini - 1
                        i_step = -1
                    for fibid in range(i_start, i_stop, i_step):
                        fiblabel_ori = fibid_with_box[fibid - 1]
                        fiblabel_new = fibid_with_box[fibid - 1 + fibid_shift]
                        if args.verbose:
                            print('(renumber_fibids) fibid:',
                                  fiblabel_ori, '-->', fiblabel_new)
                        bigdict['contents'][fibid -1 + fibid_shift] = \
                            deepcopy(bigdict['contents'][fibid -1])
                        bigdict['contents'][fibid -1 + fibid_shift]['fibid'] += \
                            fibid_shift
                        # display updated trace
                        coeff = \
                            bigdict['contents'][fibid -1 + fibid_shift]['fitparms']
                        start = \
                            bigdict['contents'][fibid -1 + fibid_shift]['start']
                        stop = bigdict['contents'][fibid -1 + fibid_shift]['stop']
                        plot_trace(ax, coeff, start, stop, ix_offset,
                                   args.rawimage, args.fibids,
                                   fiblabel_ori + '-->' + fiblabel_new,
                                   colour='green')
                    if fibid_shift == -1:
                        bigdict['contents'][fibid_end - 1]['fitparms'] = []
                    else:
                        bigdict['contents'][fibid_ini - 1]['fitparms'] = []
                else:
                    raise ValueError('fibid_shift in operation '
                                     'renumber_fibids_within_box '
                                     'must be -1 or 1')
            else:
                raise ValueError('Unexpected healing method:',
                                 operation['description'])

# update trace map
    if args.updated_traces is not None:
        # avoid overwritting initial JSON file
        if args.updated_traces.name != args.traces_file.name:
            # new random uuid for the updated calibration
            bigdict['uuid'] = str(uuid4())
            with open(args.updated_traces.name, 'w') as outfile:
                json.dump(bigdict, outfile, indent=2)

    if pdf is not None:
        pdf.savefig()
        pdf.close()
    else:
        pause_debugplot(12, pltshow=True, tight_layout=True)


if __name__ == "__main__":

    main()



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Combination routines"""

from __future__ import division

import datetime
import logging
import uuid

from astropy.io import fits

from numina.array import combine

import megaradrp.datamodel


def basic_processing_with_combination(rinput, flow,
                                      method=combine.mean,
                                      errors=True,
                                      prolog=None):
    return basic_processing_with_combination_frames(rinput.obresult.frames,
                                                    flow, method=method,
                                                    errors=errors,
                                                    prolog=prolog)


def basic_processing_with_combination_frames(frames,
                                             flow,
                                             method=combine.mean,
                                             errors=True,
                                             prolog=None
                                             ):

    _logger = logging.getLogger(__name__)
    odata = []
    cdata = []

    datamodel = megaradrp.datamodel.MegaraDataModel()
    try:
        _logger.info('processing input images')
        for frame in frames:
            hdulist = frame.open()
            fname = datamodel.get_imgid(hdulist)
            _logger.info('input is %s', fname)
            final = flow(hdulist)
            _logger.debug('output is input: %s', final is hdulist)
            cdata.append(final)
            # Files to be closed at the end
            odata.append(hdulist)
            if final is not hdulist:
                odata.append(final)

        first_image = cdata[0]
        base_header = first_image[0].header.copy()
        cnum = len(cdata)
        _logger.info("stacking %d images using '%s'", cnum, method.__name__)
        data = method([d[0].data for d in cdata], dtype='float32')
        hdu = fits.PrimaryHDU(data[0], header=base_header)
        _logger.debug('update result header')
        if prolog:
            _logger.debug('write prolog')
            hdu.header['history'] = prolog
        hdu.header['history'] = "Combined %d images using '%s'" % (cnum, method.__name__)
        hdu.header['history'] = 'Combination time {}'.format(datetime.datetime.utcnow().isoformat())
        for img in cdata:
            hdu.header['history'] = "Image {}".format(datamodel.get_imgid(img))
        prevnum = base_header.get('NUM-NCOM', 1)
        hdu.header['NUM-NCOM'] = prevnum * cnum
        hdu.header['UUID'] = str(uuid.uuid1())

        # Copy extensions and then append 'variance' and 'map'
        result = fits.HDUList([hdu])
        for hdu in first_image[1:]:
            result.append(hdu.copy())

        # Headers of last image
#        hdu.header['TSUTC2'] = cdata[-1][0].header['TSUTC2']
        # Append error extensions
        if errors:
            varhdu = fits.ImageHDU(data[1], name='VARIANCE')
            result.append(varhdu)
            num = fits.ImageHDU(data[2].astype('int16'), name='MAP')
            result.append(num)

    finally:
        _logger.debug('closing images')
        for hdulist in odata:
            hdulist.close()

    return result


def combination_hdul(hduls, method=combine.mean, errors=True, prolog=None):

    _logger = logging.getLogger(__name__)

    # FIXME: this should be merged with basic_processing_with_combination_frames
    # FIXME: DRY

    cnum = len(hduls)
    if cnum == 0:
        raise ValueError('number of HDUList == 0')

    datamodel = megaradrp.datamodel.MegaraDataModel()

    first_image = hduls[0]
    base_header = first_image[0].header.copy()

    _logger.info("stacking %d images using '%s'", cnum, method.__name__)
    combined_data = method([d[0].data for d in hduls], dtype='float32')

    hdu = fits.PrimaryHDU(combined_data[0], header=base_header)
    _logger.debug('update result header')
    if prolog:
        _logger.debug('write prolog')
        hdu.header['history'] = prolog
    hdu.header['history'] = "Combined %d images using '%s'" % (cnum, method.__name__)
    hdu.header['history'] = 'Combination time {}'.format(datetime.datetime.utcnow().isoformat())

    for img in hduls:
        hdu.header['history'] = "Image {}".format(datamodel.get_imgid(img))

    prevnum = base_header.get('NUM-NCOM', 1)
    hdu.header['NUM-NCOM'] = prevnum * cnum
    hdu.header['UUID'] = str(uuid.uuid1())

    # Copy extensions and then append 'variance' and 'map'
    result = fits.HDUList([hdu])
    for hdu in first_image[1:]:
        result.append(hdu.copy())

    # Headers of last image
#        hdu.header['TSUTC2'] = cdata[-1][0].header['TSUTC2']
    # Append error extensions
    if errors:
        varhdu = fits.ImageHDU(combined_data[1], name='VARIANCE')
        result.append(varhdu)
        num = fits.ImageHDU(combined_data[2].astype('int16'), name='MAP')
        result.append(num)

    return result


def main(args=None):
    import argparse
    import contextlib
    import astropy.io.fits as fits

    parser = argparse.ArgumentParser(prog='combine')
    parser.add_argument('-o', '--output', default='combined.fits')
    parser.add_argument('-e', '--errors', default=False, action='store_true')
    parser.add_argument('--method', default='mean', choices=['mean', 'median'])
    parser.add_argument('image', nargs='+')
    args = parser.parse_args(args)

    if args.method == 'mean':
        method = combine.mean
    elif args.method == 'median':
        method = combine.median
    else:
        raise ValueError('wrong method {}'.format(args.method))

    errors = args.errors
    with contextlib.nested(*[fits.open(fname) for fname in args.image]) as hduls:
        result = combination_hdul(hduls, method=method, errors=errors, prolog=None)

    result.writeto(args.output, overwrite=True)
    # with contextlib.ExitStack() as stack:
    #    hduls = [stack.enter_context(fits.open(fname)) for fname in args.image]
    #    combination_hdul(hduls, method=combine.mean, errors=False, prolog=None)

if __name__ == '__main__':

    main()


#
# Copyright 2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Some utils"""


import astropy.io.fits as fits


def add_collapsed_mos_extension(img, size=7, axis=0):
    """Add a collapsed image extension

    Parameters
    ==========
    img: astropy.io.fits.HDUList

    Returns
    =======
    astropy.io.fits.HDUList
        Updated image

    """

    ext = fits.ImageHDU(header=img[0].header, name='COLLAPSED')

    shape = img[0].data.shape
    if axis == 0:
        reshaped = img[0].data.reshape((-1, size, 1, shape[1])).mean(axis=(1, 2))
    elif axis == 1:
        reshaped = img[0].data.reshape((shape[0], size, 1, -1)).mean(axis=(1, 2))
    else:
        raise ValueError('invalid axis={}'.format(axis))

    ext.data = reshaped

    img.append(ext)

    return img



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy as np

from numina.simulation.device import HWDevice, Signal


class HemiCover(HWDevice):
    STATES = [0, 1]
    NAMES = {'PARKED': 1, 'INPLACE': 0}
    CODES = {1: 'PARKED', 0: 'INPLACE'}

    def __init__(self, name, parent=None):
        self._pos = 0 # Closed
        super(HemiCover, self).__init__(name, parent=parent)

        self.changed = Signal()
        self.opened = Signal()
        self.closed = Signal()

    def set(self, pos):
        if pos not in self.STATES:
            raise ValueError('%d is not a valid state' % pos)
        if pos != self._pos: # We have to move
            self.flip()

    def open(self):
        if self._pos == 0:
            self._pos = 1
            self.changed.emit(self._pos)
            self.opened.emit()

    def close(self):
        if self._pos == 1:
            self._pos = 0
            self.changed.emit(self._pos)
            self.closed.emit()

    def flip(self):
        self._pos += 1
        self._pos %= 2
        self.changed.emit(self._pos)
        if self._pos == 1:
            self.opened.emit()
        else:
            self.closed.emit()

    @property
    def position(self):
        return self._pos

    @property
    def label(self):
        return self.CODES[self._pos]


class FullCover(HWDevice):
    # STATES = 00, 01, 10, 11 
    # FULL CLOSED, CLOSED LEFT, CLOSED RIGHT, FULL OPEN
    STATES = [0, 1, 2, 3]
    NAMES = {'UNSET': 3, 'LEFT': 2, 'RIGHT': 1, 'SET': 0}
    CODES = {3: 'UNSET', 2: 'LEFT', 1: 'RIGHT', 0: 'SET'}

    def __init__(self, name=None, parent=None):
        super(FullCover, self).__init__(name, parent=parent)
        self.left = HemiCover(name='Left', parent=self)
        self.right = HemiCover(name='Right', parent=self)

        self.changed_left = self.left.changed
        self.opened_left = self.left.opened
        self.closed_left = self.left.closed

        self.changed_right = self.right.changed
        self.opened_right = self.right.opened
        self.closed_right = self.right.closed

    def flip(self):
        self.left.flip()
        self.right.flip()
        
    def open(self):
        self.left.open()
        self.right.open()
        
    def close(self):
        self.left.close()
        self.right.close()

    def set(self, pos):
        if pos not in self.STATES:
            raise ValueError('%d is not a valid state' % pos)

        l_pos = pos // 2
        r_pos = pos % 2
        self.left.set(l_pos)
        self.right.set(r_pos)

    def set_mode(self, mode):
        pos = self.NAMES[mode]
        self.set(pos)

    @property
    def position(self):
        return 2 * self.left.position + self.right.position

    @position.setter
    def position(self, pos):
        self.set(pos)

    @property
    def label(self):
        return self.CODES[self.position]

    @label.setter
    def label(self, value):
        self.set_mode(value)


class MegaraCover(FullCover):
    """MEGARA Cover"""
    NAMES = {'UNSET': 3, 'LEFT': 2, 'RIGHT': 1, 'SET': 0}
    CODES = {3: 'UNSET', 2: 'LEFT', 1: 'RIGHT', 0: 'SET'}
    HCODES = {1: 'PARKED', 0: 'INPLACE'}
  
    VALS = {3: lambda pos: np.ones(pos[:,0].shape, dtype='bool'),
            2: lambda pos: pos[:,0]<-0.7,
            1: lambda pos: pos[:,0]>0.7,
            0: lambda pos: np.zeros(pos[:,0].shape, dtype='bool')}

    def __init__(self, parent=None):
        self.mode = 'UNSET'
        super(MegaraCover, self).__init__(name='Cover', parent=parent)
        self.set_mode(self.mode)

        self._filter_s = lambda x: True
        self._cover_s = lambda x: 1.0

    def set_mode(self, mode):
        """Cover in the focal plane."""
        if mode.upper() not in self.NAMES:
            raise ValueError('"%s" mode is not recognized' % mode)

        pos = self.NAMES[mode.upper()]
        super(MegaraCover, self).set(pos)

        assert pos == self.position

        self._cover_s = lambda x: 1.0

        if self.position == 3:
            self._filter_s = lambda x: True
            self._cover_s = lambda x: 1.0
        elif self.position == 2:
            self._filter_s = lambda pos: pos[0] >= 0.0
            self._cover_s = lambda pos: 1.0 if pos[0] > 0.0 else 0.5
        elif self.position == 1:
            self._filter_s = lambda pos: pos[0] <= 0.0
            self._cover_s = lambda pos: 1.0 if pos[0] < 0.0 else 0.5
        else:
            self._filter_s = lambda pos: False
            self._cover_s = lambda pos: 0.0

    def visible_fibers(self, fibid, allpos):
        p1 = [(fid, pos[0], pos[1], self._cover_s(pos)) for fid, pos in zip(fibid, allpos) if self._filter_s(pos)]
        return p1

    def __call__(self, fiber_positions):
        return self.VALS[self.position](fiber_positions)



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Simple monocromatic simulation"""

import uuid
import math
from datetime import datetime

import astropy.wcs as wcs
import astropy.io.fits as fits
from numina.simulation.factory import extract, extractm


class MegaraImageFactory(object):
    CARDS_P = [
        ('OBSERVAT', 'ORM', 'Name of observatory'),
        ('TELESCOP', 'GTC', 'Telescope id.'),
        ('INSTRUME', 'MEGARA', 'Name of the Instrument'),
        ('ORIGIN', 'SIMULATOR_B', 'FITS file originator'),
        ('OSFILTER', False, 'Sort order filter'),
        ('INSCONF', '66f2283e-3049-4d4b-8ef1-14d62fcb611d')
    ]

    def __init__(self):
        pass

    def bun_fib_lcb(self, meta, hdr):

        sky_bundles_in_lcb = [93, 94, 95, 96, 97, 98, 99, 100]

        extract(hdr, meta, ['MEGARA.LCB', 'nfibers'], 'NFIBERS')
        extract(hdr, meta, ['MEGARA.LCB', 'nbundles'], 'NBUNDLES')
        extract(hdr, meta, ['MEGARA.LCB', 'conf_id'], 'CONFID')
        extract(hdr, meta, ['MEGARA.LCB', 'name'], 'INSMODE')

        # insert WCS
        w = wcs.WCS(naxis=2)
        w.wcs.crpix = [0, 0]
        w.wcs.ctype = ['RA---TAN', 'DEC--TAN']
        w.wcs.cdelt = [1.0 / 3600.0, 1.0 / 3600.0]
        w.wcs.crval = [0.0, 70.0]
        # Rotation around (0,0)
        ang = 1.0 / 180 * math.pi
        cs = math.cos(ang)
        ss = math.sin(ang)
        w.wcs.pc = [[cs, -ss], [ss, cs]]
        # Final rotation to a differente center
        # w.wcs.pv = [[], []]
        hdrwcs = w.to_header()
        hdr.extend(hdrwcs.cards)

        fibers_info = extractm(meta, ['MEGARA.LCB', 'fibers'])
        # FIXME: inactive
        inactive_fibs_id = []
        written_bunds = []
        for fiber_info in fibers_info:
            fibid, bunid, x, y, inactive = fiber_info

            if bunid not in written_bunds:
                written_bunds.append(bunid)

                key = "BUN%03d_P" % bunid
                hdr[key] = 0  # priority
                key = "BUN%03d_I" % bunid
                hdr[key] = "unknown"
                key = "BUN%03d_T" % bunid
                if bunid in sky_bundles_in_lcb:
                    hdr[key] = "SKY"
                else:
                    hdr[key] = "UNASSIGNED"

            key = "FIB%03d_B" % fibid
            hdr[key] = bunid
            # Coordinates
            key = "FIB%03d_D" % fibid  # DEC
            hdr[key] = 0.0000
            key = "FIB%03d_R" % fibid  # RA
            hdr[key] = 0.0000
            key = "FIB%03d_O" % fibid  # PA
            hdr[key] = 0.0000

            key = "FIB%03d_A" % fibid  # Active
            if inactive == 1:
                hdr[key] = False
            else:
                hdr[key] = True

            # Coordinates
            key = "FIB%03d_X" % fibid  # X
            hdr[key] = x
            key = "FIB%03d_Y" % fibid  # Y
            hdr[key] = y

        return hdr

    def bun_fib_mos(self, meta, hdr):

        nbundles = extractm(meta, ['MEGARA.MOS', 'nbundles'])

        extract(hdr, meta, ['MEGARA.MOS', 'nfibers'], 'NFIBERS')
        extract(hdr, meta, ['MEGARA.MOS', 'nbundles'], 'NBUNDLES')
        extract(hdr, meta, ['MEGARA.MOS', 'conf_id'], 'CONFID')
        extract(hdr, meta, ['MEGARA.MOS', 'name'], 'INSMODE')

        for i in range(1, nbundles + 1):
            rbpath = 'MEGARA.MOS.RoboticPositioner_%d' % i
            extract(hdr, meta, [rbpath, 'target_priority'], "BUN%03d_P" % i, default=0)
            extract(hdr, meta, [rbpath, 'target_name'], "BUN%03d_I" % i, default="unknown")
            extract(hdr, meta, [rbpath, 'target_type'], "BUN%03d_T" % i, default="UNASSIGNED")

            fibs_id = extractm(meta, [rbpath, 'bundle', 'fibs_id'])
            inactive_fibs_id = extractm(meta, [rbpath, 'bundle', 'inactive_fibs_id'])
            pos_fibs = extractm(meta, [rbpath, 'fibers_pos'])
            for fibid, pos in zip(fibs_id, pos_fibs):
                extract(hdr, meta, [rbpath, 'id'], "FIB%03d_B" % fibid)
                # Coordinates
                key = "FIB%03d_D" % fibid # DEC
                hdr[key] = 0.0000
                key = "FIB%03d_R" % fibid # RA
                hdr[key] = 0.0000
                key = "FIB%03d_O" % fibid # PA
                hdr[key] = 0.0000

                key = "FIB%03d_A" % fibid # Active
                if fibid in inactive_fibs_id:
                    hdr[key] = False
                else:
                    hdr[key] = True

                # Coordinates
                key = "FIB%03d_X" % fibid # X
                hdr[key] = pos[0]
                key = "FIB%03d_Y" % fibid # Y
                hdr[key] = pos[1]

        return hdr

    def create_from_instrument(self, data, name, instrument, mode=''):
        meta = instrument.config_info()
        pheader = fits.Header(self.CARDS_P)

        # pheader['FILENAME'] = name
        # OBS mode
        pheader['OBSMODE'] = mode
        pheader['DATE-OBS'] = datetime.utcnow().isoformat()
        exptime = meta[instrument.name].get('exposed', 1.0)
        pheader['EXPTIME'] = exptime
        pheader['EXPOSED'] = exptime

        hdu1 = fits.PrimaryHDU(data, header=pheader)
        hdul = fits.HDUList([hdu1])
        return hdul

    def create(self, data, name, control):

        pheader = fits.Header(self.CARDS_P)
        # pheader['FILENAME'] = name
        pheader['OBSMODE'] = control.mode
        pheader['UUID'] = str(uuid.uuid4())
        # Date of simulation
        pheader['DATE'] = datetime.utcnow().isoformat()
        # Date of simulated observation, not set yet
        pheader['DATE-OBS'] = datetime.utcnow().isoformat()

        # Seqs
        metacontrol = control.config_info()
        extract(pheader, metacontrol, ['ob_data', 'obsid'], 'OBSID', default=0.0)
        extract(pheader, metacontrol, ['ob_data', 'repeat'], 'NNREP', default=0.0)
        extract(pheader, metacontrol, ['ob_data', 'count'], 'NNSEC', default=0.0)

        instrument = control.get('MEGARA')
        telescope = control.get('GTC')

        meta = instrument.config_info()
        meta.update(telescope.config_info())

        # not yet implemented
        # pheader['AMPLAYOU'] = "NORMAL"
        # pheader['AMPUP'] = "G"
        # pheader['AMPLOW'] = "E"
        # pheader['GAINUP'] = meta_det.get('gainup', 1.0)
        # pheader['GAINLOW'] = meta_det.get('gainlow', 1.0)
        # pheader['RONUP'] = meta_det.get('ronup', 1.0)
        # pheader['RONLOW'] = meta_det.get('ronlow', 1.0)
        extract(pheader, meta, ['MEGARA.Detector', 'exposed'], 'EXPTIME')
        extract(pheader, meta, ['MEGARA.Detector', 'exposed'], 'EXPOSED')
        extract(pheader, meta, ['MEGARA.Detector', 'vbin'], 'VBIN')
        extract(pheader, meta, ['MEGARA.Detector', 'hbin'], 'HBIN')

        extract(pheader, meta, ['MEGARA.Wheel', 'selected', 'setup'], 'VPH', default='unknown')
        pheader['VPHFWHM1'] = 0.0
        pheader['VPHFWHMC'] = 0.0
        pheader['VPHFWHM2'] = 0.0
        extract(pheader, meta, ['MEGARA.Wheel', 'selected', 'wl_range'], 'VPHWL1', selector=lambda x: x[0])
        extract(pheader, meta, ['MEGARA.Wheel', 'selected', 'wl_range'], 'VPHWLC', selector=lambda x: x[1])
        extract(pheader, meta, ['MEGARA.Wheel', 'selected', 'wl_range'], 'VPHWL2', selector=lambda x: x[2])

        extract(pheader, meta, ['MEGARA.Focus', 'focus'], 'FOCUS', default=0)
        extract(pheader, meta, ['GTC.Focus', 'focus'], 'FOCUST', default=3)
        extract(pheader, meta, ['MEGARA.Cover', 'label'], 'cover')
        extract(pheader, meta, ['MEGARA.Cover.Left', 'label'], 'cover1')
        extract(pheader, meta, ['MEGARA.Cover.Right', 'label'], 'cover2')
        extract(pheader, meta, ['MEGARA', 'insmode'], 'insmode', default='unknown')

        #self.bun_fib(mode, meta, data, pheader)

        calibration_unit = control.get('ICM-MEGARA')
        meta_megcalib = calibration_unit.config_info()
        if control.mode in ['arc', 'fiberflat', 'focus']:
            # LAMP keywords
            extract(pheader, meta_megcalib, ['ICM-MEGARA', 'label'], 'LAMP')

        hdu1 = fits.PrimaryHDU(data, header=pheader)

        # Bundles and fibers
        # IN a second extension
        hdu2 = fits.ImageHDU()

        hdu2.header['EXTNAME'] = 'FIBERS'
        if pheader['insmode'] == 'MOS':
            self.bun_fib_mos(meta, hdu2.header)
        else:
            self.bun_fib_lcb(meta, hdu2.header)

        hdul = fits.HDUList([hdu1, hdu2])
        return hdul



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import pytest

import numpy
import astropy.io.fits as fits

from megaradrp.utils import add_collapsed_mos_extension


def test_add_add_collapsed_mos_extension0():

    img = fits.HDUList()

    data = numpy.ones((56, 1945), dtype='int16')
    for i in range(56):
        data[i] = i

    img.append(fits.ImageHDU(data))
    img.append(fits.ImageHDU(name='FIBERS'))

    result = add_collapsed_mos_extension(img)

    assert img[0].shape == (56, 1945)
    assert 'FIBERS' in result
    assert 'COLLAPSED' in result
    ext = result['COLLAPSED'].data
    assert ext.shape == (8, 1945)
    for i in range(7):
        assert numpy.allclose(ext[i, :], 3.0 + 7 * i)


def test_add_add_collapsed_mos_extension1():

    img = fits.HDUList()
    size0 = 1945

    size1 = 56
    size = 7
    ncol = size1 // size

    data = numpy.ones((size0, size1), dtype='int16')
    for i in range(size1):
        data[:, i] = i

    img.append(fits.ImageHDU(data))
    img.append(fits.ImageHDU(name='FIBERS'))

    result = add_collapsed_mos_extension(img, axis=1)

    assert img[0].shape == (size0, size1)
    assert 'FIBERS' in result
    assert 'COLLAPSED' in result
    ext = result['COLLAPSED'].data
    assert ext.shape == (size0, ncol)
    for i in range(size):
        assert numpy.allclose(ext[:, i], 24.0 + i)


@pytest.mark.parametrize("axis", [-1, 2])
def test_add_add_collapsed_mos_extension2(axis):

    img = fits.HDUList()

    data = numpy.ones((56, 1945), dtype='int16')
    for i in range(56):
        data[i] = i

    img.append(fits.ImageHDU(data))
    img.append(fits.ImageHDU(name='FIBERS'))

    with pytest.raises(ValueError):
        add_collapsed_mos_extension(img, axis=axis)



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Calibration Recipes for Megara"""


from numina.core import Product, Requirement
from numina.core.types import ListOfType

import megaradrp.types as typs
from megaradrp.core.recipe import MegaraBaseRecipe


class Recipe(MegaraBaseRecipe):
    """Process Sensitivity Star Recipe.

    This recipe processes a set of images
    processed by the recipes of
    **Standard star with the FIBER MOS** or
    **Standard star with the LCB IFU**
    and returns the sensitivity correction
    and the atmospheric extinction required
    for flux calibration.

    See Also
    --------
    megaradrp.recipes.combined.sensstar.Recipe

    """

    reference_spectra = Requirement(ListOfType(typs.ReferenceSpectrumTable),
                                    "Reference spectra of Std stars")

    master_sensitivity = Product(typs.MasterSensitivity)
    master_extinction = Product(typs.ReferenceExtinctionTable,
                                "Atmospheric extinction")

    def run(self, rinput):

        self.logger.info('starting ExtinctionStarRecipe reduction')

        result = super(Recipe,self).run(rinput)

        return self.create_result(final=result[0], target=result[1], sky=result[2])



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

import numpy as np
from scipy.stats import norm
import scipy.interpolate as ii
from scipy.ndimage.filters import convolve1d
from numina.simulation.device import HWDevice
from numina.simulation.efficiency import Efficiency


class InternalOptics(object):
    def __init__(self, transmission=None):

        if transmission is None:
            self._transmission = Efficiency()
        else:
            self._transmission = transmission

    def transmission(self, wl):
        return self._transmission.response(wl)


class FocusActuator(HWDevice):
    def __init__(self, name, parent=None):
        super(FocusActuator, self).__init__(name, parent)

        # Focus
        self.internal_focus_factor = 1.0
        self._ref_focus = 1000.0
        self._internal_focus = self._ref_focus

    def set_focus(self, x):
        """Arbitrary parametrization of the focus"""
        if x < -1000 or x > 3000:
            raise ValueError('focus out of limits')

        self.internal_focus_factor = 1+ 1.9 * (math.cosh((x - self._ref_focus) / 3000.0) - 1)
        self._internal_focus = x

    @property
    def focus(self):
        return self._internal_focus

    @focus.setter
    def focus(self, value):
        self.set_focus(value)


class MegaraInstrument(HWDevice):
    def __init__(self, focal_plane, pseudo_slit, internal_optics, wheel, detector, shutter):

        super(MegaraInstrument, self).__init__('MEGARA')

        self._mode = 'LCB'
        self.detector = detector
        self.detector.set_parent(self)

        self.pseudo_slit_sel = pseudo_slit
        self.pseudo_slit_sel.set_parent(self)

        self.pseudo_slit = self.pseudo_slit_sel.current()
        self.fiberset = self.pseudo_slit.fiberset

        self.focal_plane = focal_plane

        # pseudo slit and fiber bundle for current mode

        self.wheel = wheel
        self.wheel.set_parent(self)
        self.vph = self.wheel.current()

        self.internal_optics = internal_optics

        self.dev_shutter = shutter
        self.dev_shutter.set_parent(self)

        self.focal_actuator = FocusActuator('Focus', self)

        # Callbacks get called on predefined events on the devices
        # A callback to maintain self.vph updated
        # whenever the wheel is moved

        def update_current_vph(_):
            self.vph = self.wheel.current()

        self.wheel.changed.connect(update_current_vph)

        def update_current_pslit(_):
            self.pseudo_slit = self.pseudo_slit_sel.current()

        self.pseudo_slit_sel.changed.connect(update_current_pslit)

    def set_mode(self, mode):
        """Set overall mode of the instrument."""
        mode_u = mode.upper()
        if mode_u not in ['MOS', 'LCB']:
            raise ValueError('mode "%s" not valid' % (mode, ))

        self._mode = mode_u
        self.pseudo_slit_sel.select(self._mode)
        self.pseudo_slit = self.pseudo_slit_sel.current()
        self.fiberset = self.pseudo_slit.fiberset

    def set_vph(self, vphname):
        """Set VPH of the instrument."""
        self.wheel.select(vphname)

    def set_cover(self, status):
        """Cover in the focal plane."""

        # This should be in other object
        self.focal_plane.set_cover(status)
        # Update visible fibers in the ps_slit

    def set_focus(self, x):
        """Arbitrary parametrization of the focus"""

        self.focal_actuator.set_focus(x)

    def get_visible_fibers(self):
        return self.focal_plane.get_visible_fibers(self.insmode)

    @property
    def insmode(self):
        return self._mode

    @insmode.setter
    def insmode(self, value):
        self.set_mode(value)

    @property
    def cover(self):
        return self.focal_plane.cover.label

    @cover.setter
    def cover(self, state):
        self.set_cover(state)

    @property
    def shutter(self):
        return self.dev_shutter.label

    @shutter.setter
    def shutter(self, value):
        self.dev_shutter.label = value

    def project_rss_intl(self, sigma, wl_in, spec_in):

        # This will compute only the illuminated fibers
        tab = self.get_visible_fibers()
        visible_fib_ids = tab['fibid']
        return project_rss(visible_fib_ids, self.pseudo_slit, self.vph, self.detector, sigma, wl_in, spec_in)

    def project_rss(self, wl_in, spec_in):
        return self.project_rss_intl(self.focal_actuator.internal_focus_factor * self.fiberset.sigma, wl_in, spec_in)

    def project_rss_w(self):
        # This will compute only the illuminated fibers
        visible_fib_ids = [fibid for fibid, _ in self.focal_plane.get_visible_fibers()]
        sigma = self.fiberset.sigma
        return project_rss_w(visible_fib_ids, self.pseudo_slit, self.vph, sigma)

    def apply_transmissions(self, wltable_in, photons_in):
        """Simulate the image in the focal plane,"""

        spec_in = photons_in * self.fiberset.transmission(wltable_in)

        # Spectrograph optics transmission
        spec_in *= self.internal_optics.transmission(wltable_in)

        # VPH transmission
        spec_in *= self.vph.transmission(wltable_in)

        # QE of the detector
        spec_in *= self.detector.qe_wl(wltable_in)

        return self.project_rss(wltable_in, spec_in)

    def apply_transmissions_only(self, wltable_in, photons_in):
        """Simulate the image in the focal plane,"""

        spec_in = photons_in * self.fiberset.transmission(wltable_in)

        # Spectrograph optics transmission
        spec_in *= self.internal_optics.transmission(wltable_in)

        # VPH transmission
        spec_in *= self.vph.transmission(wltable_in)

        # QE of the detector
        spec_in *= self.detector.qe_wl(wltable_in)

        return spec_in

    def simulate_focal_plane(self, wltable_in, photons_in):
        """Simulate the image in the focal plane,"""

        # Input spectra in each fiber, in photons
        # Fiber flux is 0 by default
        base_coverage = np.zeros((self.fiberset.nfibers, 1))

        tab = self.get_visible_fibers()
        fibid = tab['fibid']

        base_coverage[fibid-1, 0] = tab['cover']
        
        # This should work with 1D and 2D
        photon_cover = base_coverage * photons_in

        spec_in = photon_cover * self.fiberset.transmission(wltable_in)

        # Spectrograph optics transmission
        spec_in *= self.internal_optics.transmission(wltable_in)

        # VPH transmission
        spec_in *= self.vph.transmission(wltable_in)

        # QE of the detector
        spec_in *= self.detector.qe_wl(wltable_in)

        return self.project_rss(self.focal_actuator.internal_focus_factor* self.fiberset.sigma, wltable_in, spec_in)

    def illumination_in_focal_plane(self, photons, illumination=None):

        # Input spectra in each fiber, in photons
        # Fiber flux is 1 by default
        base_coverage = np.ones((self.fiberset.nfibers, 1))

        tab = self.focal_plane.get_all_fibers(self.fiberset)

        fibid = tab['fibid']
        pos_x = tab['x']
        pos_y = tab['y']

        if illumination:
            base_coverage[fibid-1, 0] = illumination(pos_x, pos_y)
        return base_coverage * photons


def project_rss(vis_fibs_id, pseudo_slit, vph, detector, sigma, wl_in, spec_in, scale=8):

    # Scale for super sampling
    # scale 8 by default
    y_ps_fibers = pseudo_slit.y_pos(vis_fibs_id)
    DSHAPE = detector.dshape
    PIXSCALE = detector.pixscale
    # Scale for super sampling
    xcenter = detector.dshape[1] // 2
    ycenter = detector.dshape[0] // 2

    spos = PIXSCALE * (np.arange(0, DSHAPE[1] * scale) - scale * xcenter) / scale

    wl_in_super = vph.ps_x_wl(y_ps_fibers, spos, grid=True)
    # revert
    if wl_in_super[0,0] > wl_in_super[0,-1]:
        reversed = True
    else:
        reversed = False

    # print('wl in super?', wl_in_super.shape)
    spec_in_super = np.zeros_like(wl_in_super)

    # Resample to higher spatial resolution
    for i, fib_id in enumerate(vis_fibs_id):
        idx = fib_id - 1
        interpolator = ii.interp1d(wl_in, spec_in[idx]) # This is not conserving flux?
        spec_in_super[i] = interpolator(wl_in_super[i])

    # kernel is constant in pixels
    # This is a gaussian convolved with a square
    kernel = compute_kernel(scale*sigma, truncate=5.0, d=0.5 * scale)
    out = convolve1d(spec_in_super, kernel, axis=1)

    # Downsample after convolution
    spec_in_detector = out[:,::scale]
    wl_in_detector = wl_in_super[:,::scale]

    # fits.writeto('downsampled.fits', spec_in_detector, overwrite=True)
    ytrace = np.zeros_like(wl_in_detector)

    # Y-positions of the traces
    for idx, y_ps_fiber in enumerate(y_ps_fibers):
        if reversed:
            wl_i = wl_in_detector[idx, ::-1]
            # ps_wl_y inputs must be sorted
            compy = vph.ps_wl_y(y_ps_fiber, wl_i)[::-1]
        else:
            wl_i = wl_in_detector[idx]
            compy = vph.ps_wl_y(y_ps_fiber, wl_i)

        ytrace[idx] = ycenter + compy / PIXSCALE

    nsig = 6 # At 6 sigma, the value of the profile * 60000 counts is
             # << 1
    final = np.zeros(DSHAPE)

    for idx, _ in enumerate(y_ps_fibers):
        minp = coor_to_pix(ytrace[idx].min() - nsig * sigma)
        maxp = coor_to_pix(ytrace[idx].max() + nsig * sigma)
        yp = np.arange(minp, maxp)
        # Scale of the trace of width sigma
        base = pixcont_int_pix(yp[:,np.newaxis], ytrace[idx, :], sigma)
        #print base.sum(axis=0).max()
        yspec = base * spec_in_detector[idx]
        #print yspec.sum(axis=0).max(), spec_in_detector[idx].max()
        final[minp:maxp,:] += yspec

    return final


def project_rss_w(visible_fib_ids, pseudo_slit, vph, detector, sigma):
    DSHAPE = detector.dshape
    PIXSCALE = detector.pixscale

    xcenter = DSHAPE[1] // 2
    ycenter = DSHAPE[0] // 2

    y_ps_fibers = pseudo_slit.y_pos(visible_fib_ids)

    spos = -PIXSCALE * (np.arange(0, DSHAPE[1]) - xcenter)
    wl_in_detector = vph.ps_x_wl(y_ps_fibers, spos, grid=True)
    # revert
    wl_in_detector = wl_in_detector[:,::-1]

    ytrace = np.zeros_like(wl_in_detector)

    for idx, y_pslit in enumerate(y_ps_fibers):
        ytrace[idx] = ycenter + vph.b(y_pslit, wl_in_detector[idx]) / PIXSCALE

    nsig = 6 # At 6 sigma, the value of the profile * 60000 counts is
             # << 1
    fvalue = 5e-6

    l1 = []
    l2 = []
    l3 = []
    l4 = []

    for idx, _ in enumerate(y_ps_fibers):
        minp = coor_to_pix(ytrace[idx].min() - nsig * sigma)
        maxp = coor_to_pix(ytrace[idx].max() + nsig * sigma)
        yp = np.arange(minp, maxp)
        base = pixcont_int_pix(yp[:, np.newaxis], ytrace[idx, :], sigma)
        sidx = np.nonzero(base >= fvalue)
        l1.extend(minp + sidx[0])
        l2.extend(sidx[1])
        l3.extend(base[sidx])
        l4.extend([idx]*len(sidx[0]))

    return l1, l2, l3, l4
    #      row, col, val, fib

def coor_to_pix(x):
    return np.ceil(x -0.5).astype('int')


def pixcont_int_pix(i, x0, sig, d=0.5):
    """Integrate a gaussian profile."""
    zs = (i - x0) / sig
    z2 = zs + d / sig
    z1 = zs - d / sig
    return (norm.cdf(z2) - norm.cdf(z1)) / (2*d)


def pixcont_int(i, x0, sig):
    '''Integrate a gaussian profile.'''
    z2 = (i + 0.5 - x0) / sig
    z1 = (i - 0.5 - x0) / sig
    return norm.cdf(z2) - norm.cdf(z1)


def compute_kernel(sigma, d=0.5, truncate=5.0):
    """A kernel representing a Gaussian convolved with a square."""
    sd = float(sigma)
    # make the radius of the filter equal to truncate standard deviations
    lw = int(truncate * sd + 0.5)
    weights = [0.0] * (2 * lw + 1)
    weights[lw] = pixcont_int(0, 0, sigma)
    sum = weights[lw]
    # calculate the kernel:
    for ii in range(1, lw + 1):
        tmp = pixcont_int_pix(ii, 0, sigma, d)
        weights[lw + ii] = tmp
        weights[lw - ii] = tmp
        sum += 2.0 * tmp
    for ii in range(2 * lw + 1):
        weights[ii] /= sum

    return weights
